"""
    Copyright 2016 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import time

from tornado import gen
from tornado.gen import sleep


@gen.coroutine
def retry_limited(fun, timeout):
    start = time.time()
    while time.time() - start < timeout and not fun():
        yield sleep(0.1)
    if not fun():
        raise AssertionError("Bounded wait failed")


UNKWN = object()


def assert_equal_ish(minimal, actual, sortby=[]):
    if isinstance(minimal, dict):
        for k in minimal.keys():
            assert_equal_ish(minimal[k], actual[k], sortby)
    elif isinstance(minimal, list):
        assert len(minimal) == len(actual), "list not equal %s != %s" % (minimal, actual)
        if len(sortby) > 0:
            def keyfunc(val):
                if not isinstance(val, dict):
                    return val
                key = [str(val[x]) for x in sortby if x in val]
                return '_'.join(key)
            actual = sorted(actual, key=keyfunc)
        for (m, a) in zip(minimal, actual):
            assert_equal_ish(m, a, sortby)
    elif minimal is UNKWN:
        return
    else:
        assert minimal == actual


def assert_graph(graph, expected):
    lines = ["%s: %s" % (f.id.get_attribute_value(), t.id.get_attribute_value()) for f in graph.values() for t in f.requires]
    lines = sorted(lines)

    elines = [x.strip() for x in expected.split("\n")]
    elines = sorted(elines)

    assert elines == lines, (lines, elines)



from inmanta.execute.util import Unknown
from inmanta.plugins import plugin
from _collections import defaultdict


@plugin
def unknown() -> "any":
    return Unknown(None)


@plugin
def length(string: "string") -> "number":
    return len(string)


@plugin
def empty(string: "string") -> "bool":
    return len(string) == 0


@plugin(allow_unknown=True)
def is_uknown(inp: "any") -> "bool":
    return isinstance(inp, Unknown)


@plugin(allow_unknown=True)
def do_uknown(inp: "any") -> "string":
    return "XX"


counter = defaultdict(lambda: 0)


@plugin
def once(string: "string") -> "number":
    prev = counter[string]
    counter[string] = prev + 1
    return prev



"""
    Copyright 2018 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
from inmanta.util import Scheduler
from inmanta.protocol import RESTBase, decode_token, json_encode, UnauhorizedError, ReturnClient, handle

from inmanta import config as inmanta_config, methods
from inmanta.server import config as opt, SLICE_SESSION_MANAGER

import tornado.web
from tornado import gen, queues
from tornado.ioloop import IOLoop
from tornado.httpserver import HTTPServer

import logging
import ssl
import time
import uuid
from _collections import defaultdict


LOGGER = logging.getLogger(__name__)


# Server Side
class RESTServer(RESTBase):

    def __init__(self, connection_timout=120):
        self.__end_points = []
        self.__endpoint_dict = {}
        self._handlers = []
        self.token = inmanta_config.Config.get(self.id, "token", None)
        self.connection_timout = connection_timout
        self.headers = set()
        self.sessions_handler = SessionManager(IOLoop.current())
        self.add_endpoint(self.sessions_handler)

    def add_endpoint(self, endpoint: "ServerSlice"):
        self.__end_points.append(endpoint)
        self.__endpoint_dict[endpoint.name] = endpoint

    def get_endpoint(self, name):
        return self.__endpoint_dict[name]

    def validate_sid(self, sid):
        return self.sessions_handler.validate_sid(sid)

    def get_id(self):
        """
            Returns a unique id for a transport on an endpoint
        """
        return "server_rest_transport"

    id = property(get_id)

    def create_op_mapping(self):
        """
            Build a mapping between urls, ops and methods
        """
        url_map = defaultdict(dict)

        # TODO: avoid colliding handlers

        for endpoint in self.__end_points:
            for method, method_handlers in endpoint.__methods__.items():
                properties = method.__protocol_properties__
                call = (endpoint, method_handlers[0])

                if "arg_options" in properties:
                    for opts in properties["arg_options"].values():
                        if "header" in opts:
                            self.headers.add(opts["header"])

                url = self._create_base_url(properties)
                properties["api_version"] = "1"
                url_map[url][properties["operation"]] = (properties, call, method.__wrapped__)
                url = self._create_base_url(properties, versioned=False)
                properties = properties.copy()
                properties["api_version"] = None
                url_map[url][properties["operation"]] = (properties, call, method.__wrapped__)
        return url_map

    def start(self):
        """
            Start the transport
        """
        LOGGER.debug("Starting Server Rest Endpoint")

        for endpoint in self.__end_points:
            endpoint.prestart(self)

        for endpoint in self.__end_points:
            endpoint.start()
            self._handlers.extend(endpoint.get_handlers())

        url_map = self.create_op_mapping()

        for url, configs in url_map.items():
            handler_config = {}
            for op, cfg in configs.items():
                handler_config[op] = cfg

            self._handlers.append((url, RESTHandler, {"transport": self, "config": handler_config}))
            LOGGER.debug("Registering handler(s) for url %s and methods %s" % (url, ", ".join(handler_config.keys())))

        port = 8888
        if self.id in inmanta_config.Config.get() and "port" in inmanta_config.Config.get()[self.id]:
            port = inmanta_config.Config.get()[self.id]["port"]

        application = tornado.web.Application(self._handlers, compress_response=True)

        crt = inmanta_config.Config.get("server", "ssl_cert_file", None)
        key = inmanta_config.Config.get("server", "ssl_key_file", None)

        if(crt is not None and key is not None):
            ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            ssl_ctx.load_cert_chain(crt, key)

            self.http_server = HTTPServer(application, decompress_request=True, ssl_options=ssl_ctx)
            LOGGER.debug("Created REST transport with SSL")
        else:
            self.http_server = HTTPServer(application, decompress_request=True)

        self.http_server.listen(port)

        LOGGER.debug("Start REST transport")

    def stop(self):
        LOGGER.debug("Stoppin Server Rest Endpoint")
        self.http_server.stop()
        for endpoint in self.__end_points:
            endpoint.stop()

    def return_error_msg(self, status=500, msg="", headers={}):
        body = {"message": msg}
        headers["Content-Type"] = "application/json"
        LOGGER.debug("Signaling error to client: %d, %s, %s", status, body, headers)
        return body, headers, status


class ServerSlice(object):
    """
        An API serving part of the server.
    """

    def __init__(self, io_loop, name):
        self._name = name
        self._io_loop = io_loop

        self.create_endpoint_metadata()
        self._end_point_names = []
        self._handlers = []
        self._sched = Scheduler(self._io_loop)

    def prestart(self, server: RESTServer):
        """Called by the RestServer host prior to start, can be used to collect references to other server slices"""
        pass

    def start(self):
        pass

    def stop(self):
        pass

    name = property(lambda self: self._name)

    def get_handlers(self):
        return self._handlers

    def get_end_point_names(self):
        # TODO: why?
        return self._end_point_names

    def add_end_point_name(self, name):
        """
            Add an additional name to this endpoint to which it reacts and sends out in heartbeats
        """
        LOGGER.debug("Adding '%s' as endpoint", name)
        self._end_point_names.append(name)

    def add_future(self, future):
        """
            Add a future to the ioloop to be handled, but do not require the result.
        """
        def handle_result(f):
            try:
                f.result()
            except Exception as e:
                LOGGER.exception("An exception occurred while handling a future: %s", str(e))

        self._io_loop.add_future(future, handle_result)

    def schedule(self, call, interval=60):
        self._sched.add_action(call, interval)

    def create_endpoint_metadata(self):
        total_dict = {method_name: getattr(self, method_name)
                      for method_name in dir(self) if callable(getattr(self, method_name))}

        methods = {}
        for name, attr in total_dict.items():
            if name[0:2] != "__" and hasattr(attr, "__protocol_method__"):
                if attr.__protocol_method__ in methods:
                    raise Exception("Unable to register multiple handlers for the same method. %s" % attr.__protocol_method__)

                methods[attr.__protocol_method__] = (name, attr)

        self.__methods__ = methods

    def add_static_handler(self, location, path, default_filename=None, start=False):
        """
            Configure a static handler to serve data from the specified path.
        """
        if location[0] != "/":
            location = "/" + location

        if location[-1] != "/":
            location = location + "/"

        options = {"path": path}
        if default_filename is None:
            options["default_filename"] = "index.html"

        self._handlers.append((r"%s(.*)" % location, tornado.web.StaticFileHandler, options))
        self._handlers.append((r"%s" % location[:-1], tornado.web.RedirectHandler, {"url": location}))

        if start:
            self._handlers.append((r"/", tornado.web.RedirectHandler, {"url": location}))

    def add_static_content(self, path, content, content_type="application/javascript"):
        self._handlers.append((r"%s(.*)" % path, StaticContentHandler, {"transport": self, "content": content,
                                                                        "content_type": content_type}))


class Session(object):
    """
        An environment that segments agents connected to the server
    """

    def __init__(self, sessionstore, io_loop, sid, hang_interval, timout, tid, endpoint_names, nodename):
        self._sid = sid
        self._interval = hang_interval
        self._timeout = timout
        self._sessionstore = sessionstore
        self._seen = time.time()
        self._callhandle = None
        self.expired = False

        self.tid = tid
        self.endpoint_names = endpoint_names
        self.nodename = nodename

        self._io_loop = io_loop

        self._replies = {}
        self.check_expire()
        self._queue = queues.Queue()

        self.client = ReturnClient(str(sid), self)

    def check_expire(self):
        if self.expired:
            LOGGER.exception("Tried to expire session already expired")
        ttw = self._timeout + self._seen - time.time()
        if ttw < 0:
            self.expire(self._seen - time.time())
        else:
            self._callhandle = self._io_loop.call_later(ttw, self.check_expire)

    def get_id(self):
        return self._sid

    id = property(get_id)

    def expire(self, timeout):
        self.expired = True
        if self._callhandle is not None:
            self._io_loop.remove_timeout(self._callhandle)
        self._sessionstore.expire(self, timeout)

    def seen(self):
        self._seen = time.time()

    def _set_timeout(self, future, timeout, log_message):
        def on_timeout():
            if not self.expired:
                LOGGER.warning(log_message)
            future.set_exception(gen.TimeoutError())

        timeout_handle = self._io_loop.add_timeout(self._io_loop.time() + timeout, on_timeout)
        future.add_done_callback(lambda _: self._io_loop.remove_timeout(timeout_handle))

    def put_call(self, call_spec, timeout=10):
        future = tornado.concurrent.Future()

        reply_id = uuid.uuid4()

        LOGGER.debug("Putting call %s: %s %s for agent %s in queue", reply_id, call_spec["method"], call_spec["url"], self._sid)

        q = self._queue
        call_spec["reply_id"] = reply_id
        q.put(call_spec)
        self._set_timeout(future, timeout, "Call %s: %s %s for agent %s timed out." %
                          (reply_id, call_spec["method"], call_spec["url"], self._sid))
        self._replies[call_spec["reply_id"]] = future

        return future

    @gen.coroutine
    def get_calls(self):
        """
            Get all calls queued for a node. If no work is available, wait until timeout. This method returns none if a call
            fails.
        """
        try:
            q = self._queue
            call_list = []
            call = yield q.get(timeout=self._io_loop.time() + self._interval)
            call_list.append(call)
            while q.qsize() > 0:
                call = yield q.get()
                call_list.append(call)

            return call_list

        except gen.TimeoutError:
            return None

    def set_reply(self, reply_id, data):
        LOGGER.log(3, "Received Reply: %s", reply_id)
        if reply_id in self._replies:
            future = self._replies[reply_id]
            del self._replies[reply_id]
            if not future.done():
                future.set_result(data)
        else:
            LOGGER.debug("Received Reply that is unknown: %s", reply_id)

    def get_client(self):
        return self.client


class SessionListener(object):

    def new_session(self, session: Session):
        pass

    def expire(self, session: Session, timeout):
        pass

    def seen(self, session: Session, endpoint_names: list):
        pass


# Internals
class SessionManager(ServerSlice):
    """
        A service that receives method calls over one or more transports
    """
    __methods__ = {}

    def __init__(self, io_loop):
        super().__init__(io_loop, SLICE_SESSION_MANAGER)

        # Config
        interval = opt.agent_timeout.get()
        hangtime = opt.agent_hangtime.get()

        if hangtime is None:
            hangtime = interval * 3 / 4

        self.hangtime = hangtime
        self.interval = interval

        # Session management
        self._heartbeat_cb = None
        self.agent_handles = {}
        self._sessions = {}

        # Listeners
        self.listeners = []

    def add_listener(self, listener):
        self.listeners.append(listener)

    def stop(self):
        """
            Stop the end-point and all of its transports
        """
        # terminate all sessions cleanly
        for session in self._sessions.copy().values():
            session.expire(0)

    def validate_sid(self, sid):
        if isinstance(sid, str):
            sid = uuid.UUID(sid)
        return sid in self._sessions

    def get_or_create_session(self, sid, tid, endpoint_names, nodename):
        if isinstance(sid, str):
            sid = uuid.UUID(sid)

        if sid not in self._sessions:
            session = self.new_session(sid, tid, endpoint_names, nodename)
            self._sessions[sid] = session
            for listener in self.listeners:
                listener.new_session(session)
        else:
            session = self._sessions[sid]
            self.seen(session, endpoint_names)
            for listener in self.listeners:
                listener.seen(session, endpoint_names)

        return session

    def new_session(self, sid, tid, endpoint_names, nodename):
        LOGGER.debug("New session with id %s on node %s for env %s with endpoints %s" % (sid, nodename, tid, endpoint_names))
        return Session(self, self._io_loop, sid, self.hangtime, self.interval, tid, endpoint_names, nodename)

    def expire(self, session: Session, timeout):
        LOGGER.debug("Expired session with id %s, last seen %d seconds ago" % (session.get_id(), timeout))
        for listener in self.listeners:
            listener.expire(session, timeout)
        del self._sessions[session.id]

    def seen(self, session: Session, endpoint_names: list):
        LOGGER.debug("Seen session with id %s" % (session.get_id()))
        session.seen()

    @handle(methods.HeartBeatMethod.heartbeat, env="tid")
    @gen.coroutine
    def heartbeat(self, sid, env, endpoint_names, nodename):
        LOGGER.debug("Received heartbeat from %s for agents %s in %s", nodename, ",".join(endpoint_names), env.id)

        session = self.get_or_create_session(sid, env.id, endpoint_names, nodename)

        LOGGER.debug("Let node %s wait for method calls to become available. (long poll)", nodename)
        call_list = yield session.get_calls()
        if call_list is not None:
            LOGGER.debug("Pushing %d method calls to node %s", len(call_list), nodename)
            return 200, {"method_calls": call_list}
        else:
            LOGGER.debug("Heartbeat wait expired for %s, returning. (long poll)", nodename)

        return 200

    @handle(methods.HeartBeatMethod.heartbeat_reply)
    @gen.coroutine
    def heartbeat_reply(self, sid, reply_id, data):
        try:
            env = self._sessions[sid]
            env.set_reply(reply_id, data)
            return 200
        except Exception:
            LOGGER.warning("could not deliver agent reply with sid=%s and reply_id=%s" % (sid, reply_id), exc_info=True)


class RESTHandler(tornado.web.RequestHandler):
    """
        A generic class use by the transport
    """

    def initialize(self, transport: "RESTServer", config):
        self._transport = transport
        self._config = config

    def _get_config(self, http_method):
        if http_method.upper() not in self._config:
            allowed = ", ".join(self._config.keys())
            self.set_header("Allow", allowed)
            self._transport.return_error_msg(405, "%s is not supported for this url. Supported methods: %s" %
                                             (http_method, allowed))
            return

        return self._config[http_method]

    def get_auth_token(self, headers: dict):
        """
            Get the auth token provided by the caller. The token is provided as a bearer token.
        """
        if "Authorization" not in headers:
            return None

        parts = headers["Authorization"].split(" ")
        if len(parts) == 0 or parts[0].lower() != "bearer" or len(parts) > 2 or len(parts) == 1:
            LOGGER.warning("Invalid authentication header, Inmanta expects a bearer token. (%s was provided)",
                           headers["Authorization"])
            return None

        return decode_token(parts[1])

    def respond(self, body, headers, status):
        if body is not None:
            self.write(json_encode(body))

        for header, value in headers.items():
            self.set_header(header, value)

        self.set_status(status)

    @gen.coroutine
    def _call(self, kwargs, http_method, call_config):
        """
            An rpc like call
        """
        if call_config is None:
            body, headers, status = self._transport.return_error_msg(404, "This method does not exist.")
            self.respond(body, headers, status)
            return

        self.set_header("Access-Control-Allow-Origin", "*")
        try:
            message = self._transport._decode(self.request.body)
            if message is None:
                message = {}

            for key, value in self.request.query_arguments.items():
                if len(value) == 1:
                    message[key] = value[0].decode("latin-1")
                else:
                    message[key] = [v.decode("latin-1") for v in value]

            request_headers = self.request.headers

            try:
                auth_token = self.get_auth_token(request_headers)
            except UnauhorizedError as e:
                self.respond(*self._transport.return_error_msg(403, "Access denied: " + e.args[0]))
                return

            auth_enabled = inmanta_config.Config.get("server", "auth", False)
            if not auth_enabled or auth_token is not None:
                result = yield self._transport._execute_call(kwargs, http_method, call_config,
                                                             message, request_headers, auth_token)
                self.respond(*result)
            else:
                self.respond(*self._transport.return_error_msg(401, "Access to this resource is unauthorized."))
        except ValueError:
            LOGGER.exception("An exception occured")
            self.respond(*self._transport.return_error_msg(500, "Unable to decode request body"))

    @gen.coroutine
    def head(self, *args, **kwargs):
        yield self._call(http_method="HEAD", call_config=self._get_config("HEAD"), kwargs=kwargs)

    @gen.coroutine
    def get(self, *args, **kwargs):
        yield self._call(http_method="GET", call_config=self._get_config("GET"), kwargs=kwargs)

    @gen.coroutine
    def post(self, *args, **kwargs):
        yield self._call(http_method="POST", call_config=self._get_config("POST"), kwargs=kwargs)

    @gen.coroutine
    def delete(self, *args, **kwargs):
        yield self._call(http_method="DELETE", call_config=self._get_config("DELETE"), kwargs=kwargs)

    @gen.coroutine
    def patch(self, *args, **kwargs):
        yield self._call(http_method="PATCH", call_config=self._get_config("PATCH"), kwargs=kwargs)

    @gen.coroutine
    def put(self, *args, **kwargs):
        yield self._call(http_method="PUT", call_config=self._get_config("PUT"), kwargs=kwargs)

    @gen.coroutine
    def options(self, *args, **kwargs):
        allow_headers = "Origin, Accept, Content-Type, X-Requested-With, X-CSRF-Token"
        if len(self._transport.headers):
            allow_headers += ", " + ", ".join(self._transport.headers)

        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Methods", "HEAD, GET, POST, PUT, OPTIONS, DELETE, PATCH")
        self.set_header("Access-Control-Allow-Headers", allow_headers)

        self.set_status(200)


class StaticContentHandler(tornado.web.RequestHandler):
    def initialize(self, transport: "RESTServer", content, content_type):
        self._transport = transport
        self._content = content
        self._content_type = content_type

    def get(self, *args, **kwargs):
        self.set_header("Content-Type", self._content_type)
        self.write(self._content)
        self.set_status(200)



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import pytest


@pytest.mark.gen_test
def test_project(server, client, cli):
    # create a new project
    result = yield cli.run("project", "create", "-n", "test_project")
    assert result.exit_code == 0

    projects = yield client.list_projects()
    assert len(projects.result["projects"]) == 1

    project = projects.result["projects"][0]
    assert project["id"] in result.output
    assert project["name"] in result.output

    # show the project
    result = yield cli.run("project", "show", project["id"])
    assert result.exit_code == 0
    assert project["id"] in result.output
    assert project["name"] in result.output

    result = yield cli.run("project", "show", project["name"])
    assert result.exit_code == 0
    assert project["id"] in result.output
    assert project["name"] in result.output

    # modify the project
    new_name = "test_project_2"
    result = yield cli.run("project", "modify", "-n", new_name, project["name"])
    assert result.exit_code == 0
    assert project["id"] in result.output
    assert new_name in result.output

    new_name = "test_project_3"
    result = yield cli.run("project", "modify", "-n", new_name, project["id"])
    assert result.exit_code == 0
    assert project["id"] in result.output
    assert new_name in result.output

    # delete the project
    result = yield cli.run("project", "delete", project["id"])
    assert result.exit_code == 0


@pytest.mark.gen_test
def test_environment(server, client, cli):
    project_name = "test_project"
    result = yield client.create_project(project_name)
    assert result.code == 200
    project_id = result.result["project"]["id"]

    # create a new environment
    result = yield cli.run("environment", "create", "-n", "test1", "-r", "/git/repo", "-b", "dev1", "-p", project_name)
    assert result.exit_code == 0

    result = yield cli.run("environment", "create", "-n", "test2", "-r", "/git/repo", "-b", "dev2", "-p", project_id)
    assert result.exit_code == 0

    environments = yield client.list_environments()
    assert len(environments.result["environments"]) == 2
    environments = environments.result["environments"]

    # list environments
    result = yield cli.run("environment", "list")
    assert result.exit_code == 0
    assert "test_project" in result.output
    assert "test1" in result.output
    assert "test2" in result.output

    # show an environment
    env_name = environments[0]["name"]
    env_id = environments[0]["id"]

    result = yield cli.run("environment", "show", env_name)
    assert result.exit_code == 0
    assert env_name in result.output
    assert env_id in result.output

    result = yield cli.run("environment", "show", env_id)
    assert result.exit_code == 0
    assert env_name in result.output
    assert env_id in result.output


@pytest.mark.gen_test
def test_environment_settings(server, environment, client, cli):
    result = yield cli.run("environment", "setting", "list", "-e", environment)
    assert result.exit_code == 0

    result = yield cli.run("environment", "setting", "set", "-e", environment, "-k", "auto_deploy", "-o", "true")
    assert result.exit_code == 0
    result = yield cli.run("environment", "setting", "set", "-e", environment, "--key", "auto_deploy", "--value", "true")
    assert result.exit_code == 0

    result = yield cli.run("environment", "setting", "list", "-e", environment)
    assert result.exit_code == 0
    assert environment in result.output
    assert "auto_deploy" in result.output

    result = yield cli.run("environment", "setting", "get", "-e", environment, "--key", "auto_deploy")
    assert result.exit_code == 0
    assert "True" in result.output

    result = yield cli.run("environment", "setting", "delete", "-e", environment, "--key", "auto_deploy")
    assert result.exit_code == 0


@pytest.mark.gen_test
def test_agent(server, client, environment, cli):
    result = yield cli.run("agent", "list", "-e", environment)
    assert result.exit_code == 0


@pytest.mark.gen_test
def test_version(server, client, environment, cli):
    version = "12345"
    resources = [{'key': 'key1',
                  'value': 'value1',
                  'id': 'test::Resource[agent1,key=key1],v=' + version,
                  'send_event': False,
                  'purged': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  'requires': ['test::Resource[agent1,key=key2],v=' + version],
                  },
                 {'key': 'key2',
                  'value': 'value2',
                  'id': 'test::Resource[agent1,key=key2],v=' + version,
                  'send_event': False,
                  'requires': [],
                  'purged': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  },
                 {'key': 'key3',
                  'value': None,
                  'id': 'test::Resource[agent1,key=key3],v=' + version,
                  'send_event': False,
                  'requires': [],
                  'purged': True,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  }
                 ]

    result = yield client.put_version(tid=environment, version=version, resources=resources, unknowns=[], version_info={})
    assert result.code == 200

    result = yield cli.run("version", "list", "-e", environment)
    assert result.exit_code == 0
    assert version in result.output
    assert "pending" in result.output

    result = yield cli.run("version", "release", "-e", environment, version)
    assert result.exit_code == 0
    assert version in result.output

    result = yield client.get_version(environment, version)
    assert result.code == 200
    assert result.result["model"]["result"] == "deploying"

    result = yield cli.run("version", "report", "-e", environment, "-i", version)
    assert result.exit_code == 0


@pytest.mark.gen_test
def test_param(server, client, environment, cli):
    result = yield cli.run("param", "set", "-e", environment, "--name", "var1", "--value", "value1")
    assert result.exit_code == 0
    assert "value1" in result.output

    result = yield cli.run("param", "get", "-e", environment, "--name", "var1")
    assert result.exit_code == 0
    assert "value1" in result.output

    result = yield cli.run("param", "list", "-e", environment)
    assert result.exit_code == 0
    assert "var1" in result.output


@pytest.mark.gen_test
def test_form_and_records(server, client, environment, cli):
    form_type = "FormType"
    result = yield client.put_form(tid=environment, id=form_type,
                                   form={'attributes': {'field1': {'default': 1, 'options': {'min': 1, 'max': 100},
                                                                   'type': 'number'},
                                                        'field2': {'default': "", 'options': {}, 'type': 'string'}},
                                         'options': {},
                                         'type': form_type}
                                   )
    assert(result.code == 200)
    form_id = result.result["form"]["id"]

    result = yield cli.run("form", "list", "-e", environment)
    assert result.exit_code == 0
    assert form_id in result.output

    result = yield cli.run("form", "show", "-e", environment, "-t", form_type)
    assert result.exit_code == 0
    assert "field1" in result.output
    assert "field2" in result.output

    result = yield cli.run("record", "create", "-e", environment, "-t", form_type, "-p", "field1=1234", "-p", "field2=test456")
    assert result.exit_code == 0
    assert "1234" in result.output
    assert "test456" in result.output

    records = yield client.list_records(tid=environment, form_type=form_type)
    assert records.code == 200
    record = records.result["records"][0]
    record_id = record["id"]

    result = yield cli.run("record", "list", "-e", environment, "-t", form_type)
    assert result.exit_code == 0
    assert record_id in result.output

    result = yield cli.run("record", "list", "-e", environment, "-t", form_type, "--show-all")
    assert result.exit_code == 0

    result = yield cli.run("record", "update", "-e", environment, "-r", record_id, "-p", "field1=98765")
    assert result.exit_code == 0
    assert "98765" in result.output

    result = yield cli.run("record", "delete", "-e", environment, record_id)
    assert result.exit_code == 0

    records = yield client.list_records(tid=environment, form_type=form_type)
    assert records.code == 200
    assert len(records.result["records"]) == 0


@pytest.mark.gen_test
def test_import_export(server, client, environment, cli, tmpdir):
    form_type = "FormType"
    result = yield client.put_form(tid=environment, id=form_type,
                                   form={'attributes': {'field1': {'default': 1, 'options': {'min': 1, 'max': 100},
                                                                   'type': 'number'},
                                                        'field2': {'default': "", 'options': {}, 'type': 'string'}},
                                         'options': {},
                                         'type': form_type}
                                   )
    form_id = result.result["form"]["id"]

    result = yield cli.run("record", "create", "-e", environment, "-t", form_type, "-p", "field1=1234", "-p", "field2=test456")
    assert result.exit_code == 0

    result = yield cli.run("form", "export", "-e", environment, "-t", form_type)
    assert form_id in result.output

    f = tmpdir.join("export.json")
    f.write(result.output)

    records = yield client.list_records(tid=environment, form_type=form_type)
    assert records.code == 200
    record = records.result["records"][0]

    yield client.delete_record(tid=environment, id=record["id"])
    records = yield client.list_records(tid=environment, form_type=form_type)
    assert records.code == 200
    assert len(records.result["records"]) == 0

    result = yield cli.run("form", "import", "-e", environment, "-t", form_type, "--file", str(f))

    records = yield client.list_records(tid=environment, form_type=form_type)
    assert records.code == 200
    assert len(records.result["records"]) == 1



"""
    Copyright 2016 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import pytest


@pytest.mark.gen_test
def test_project_api(client):
    result = yield client.create_project("project-test")
    assert result.code == 200
    assert "project" in result.result
    assert "id" in result.result["project"]

    project_id = result.result["project"]["id"]

    result = yield client.create_project("project-test")
    assert result.code == 500

    result = yield client.list_projects()
    assert result.code == 200
    assert "projects" in result.result
    assert len(result.result["projects"]) == 1

    assert result.result["projects"][0]['id'] == project_id

    result = yield client.get_project(id=project_id)
    assert result.code == 200
    assert "project" in result.result
    assert result.result["project"]['id'] == project_id
    assert result.result["project"]['name'] == "project-test"

    result = yield client.modify_project(id=project_id, name="project-test2")
    assert result.code == 200
    assert "project" in result.result
    assert result.result["project"]['id'] == project_id
    assert result.result["project"]['name'] == "project-test2"

    result = yield client.get_project(id=project_id)
    assert result.code == 200
    assert "project" in result.result
    assert result.result["project"]['id'] == project_id
    assert result.result["project"]['name'] == "project-test2"

    result = yield client.delete_project(id=project_id)
    assert result.code == 200

    result = yield client.list_projects()
    assert result.code == 200
    assert "projects" in result.result
    assert len(result.result["projects"]) == 0


@pytest.mark.gen_test
def test_env_api(client):
    result = yield client.create_project("env-test")
    assert result.code == 200
    assert "project" in result.result
    assert "id" in result.result["project"]
    project_id = result.result["project"]["id"]

    result = yield client.create_environment(project_id=project_id, name="dev")
    assert result.code == 200
    assert "environment" in result.result
    assert "id" in result.result["environment"]
    assert "project" in result.result["environment"]
    assert project_id == result.result["environment"]["project"]
    assert "dev" == result.result["environment"]["name"]

    env_id = result.result["environment"]["id"]

    result = yield client.modify_environment(id=env_id, name="dev2")
    assert result.code == 200
    assert "environment" in result.result
    assert result.result["environment"]['id'] == env_id
    assert result.result["environment"]['name'] == "dev2"

    result = yield client.get_environment(id=env_id)
    assert result.code == 200
    assert "environment" in result.result
    assert result.result["environment"]['id'] == env_id
    assert result.result["environment"]['project'] == project_id
    assert result.result["environment"]['name'] == "dev2"

    project_result = yield client.get_project(id=project_id)
    assert project_result.code == 200
    assert "project" in project_result.result
    assert env_id in project_result.result["project"]["environments"]

    result = yield client.list_environments()
    assert result.code == 200
    assert len(result.result) == 1

    result = yield client.delete_environment(id=env_id)
    assert result.code == 200

    result = yield client.list_environments()
    assert result.code == 200
    assert "environments" in result.result
    assert len(result.result["environments"]) == 0


@pytest.mark.gen_test
def test_project_cascade(client):
    result = yield client.create_project("env-test")
    project_id = result.result["project"]["id"]

    result = yield client.create_environment(project_id=project_id, name="dev")
    result = yield client.create_environment(project_id=project_id, name="prod")

    result = yield client.delete_project(project_id)
    assert result.code == 200

    result = yield client.list_environments()
    assert len(result.result["environments"]) == 0



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
# pylint: disable-msg=R0923,W0613

import logging

from . import DefinitionStatement
from inmanta.ast.type import ConstraintType, Type
from inmanta.ast.attribute import Attribute, RelationAttribute
from inmanta.ast.entity import Implementation, Entity, Default, Implement
from inmanta.ast.constraint.expression import Equals
from inmanta.ast.statements import TypeDefinitionStatement, Statement, ExpressionStatement, Literal, BiStatement
from inmanta.ast import Namespace, TypingException, DuplicateException, TypeNotFoundException, NotFoundException,\
    LocatableString, TypeReferenceAnchor, AttributeReferenceAnchor, IndexException
from typing import List
from inmanta.execute.runtime import ResultVariable, ExecutionUnit
from inmanta.ast.blocks import BasicBlock


LOGGER = logging.getLogger(__name__)


class DefineAttribute(Statement):

    def __init__(self, attr_type: LocatableString, name: LocatableString, default_value: ExpressionStatement=None,
                 multi=False, remove_default=True, nullable=False) -> None:
        """
            if default_value is None, this is an explicit removal of a default value
        """
        super(DefineAttribute, self).__init__()
        self.type = attr_type
        self.name = name
        self.default = default_value
        self.multi = multi
        self.remove_default = remove_default
        self.nullable = nullable


class DefineEntity(TypeDefinitionStatement):
    """
        Define a new entity in the configuration
    """

    def __init__(self,
                 namespace: Namespace,
                 lname: LocatableString,
                 comment: LocatableString,
                 parents: List[LocatableString],
                 attributes: List[DefineAttribute]) -> None:
        name = str(lname)
        TypeDefinitionStatement.__init__(self, namespace, name)

        self.anchors = [TypeReferenceAnchor(x.get_location(), namespace, str(x)) for x in parents]

        self.name = name
        self.attributes = attributes
        if comment is not None:
            self.comment = str(comment)
        else:
            self.comment = None

        self.parents = [str(x) for x in parents]

        if len(self.parents) == 0 and not (self.name == "Entity" and self.namespace.name == "std"):
            self.parents.append("std::Entity")

        self.type = Entity(self.name, namespace)
        self.type.location = lname.location

    def add_attribute(self, attr_type: LocatableString, name: LocatableString, default_value: ExpressionStatement=None):
        """
            Add an attribute to this entity
        """
        self.attributes.append(DefineAttribute(attr_type, name, default_value))

    def __repr__(self) -> str:
        """
            A textual representation of this entity
        """
        return "Entity(%s)" % self.name

    def get_full_parent_names(self) -> List[str]:
        try:
            return [self.namespace.get_type(str(parent)).get_full_name() for parent in self.parents]
        except TypeNotFoundException as e:
            e.set_statement(self)
            raise e

    def evaluate(self) -> None:
        """
            Evaluate this statement.
        """
        try:
            entity_type = self.type
            entity_type.comment = self.comment

            add_attributes = {}
            for attribute in self.attributes:
                attr_type = self.namespace.get_type(str(attribute.type))
                if not isinstance(attr_type, (Type, type)):
                    raise TypingException(self, "Attributes can only be a type. Entities need to be defined as relations.")

                name = str(attribute.name)
                attr_obj = Attribute(entity_type, attr_type, name, attribute.multi, attribute.nullable)
                attr_obj.location = attribute.get_location()
                self.anchors.append(TypeReferenceAnchor(
                    attribute.type.get_location(), self.namespace, str(attribute.type)))

                if name in add_attributes:
                    raise DuplicateException(
                        attr_obj,
                        add_attributes[name],
                        "Same attribute defined twice in one entity")

                add_attributes[name] = attr_obj

                if attribute.default is not None:
                    entity_type.add_default_value(name, attribute.default)
                elif attribute.remove_default:
                    entity_type.add_default_value(name, None)

            if len(set(self.parents)) != len(self.parents):
                raise TypingException(self, "same parent defined twice")
            for parent in self.parents:
                parent_type = self.namespace.get_type(str(parent))
                if parent_type is self.type:
                    raise TypingException(self, "Entity can not be its own parent (%s) " % parent)
                if not isinstance(parent_type, Entity):
                    raise TypingException(self, "Parents of an entity need to be entities. "
                                          "Default constructors are not supported. %s is not an entity" % parent)

                entity_type.parent_entities.append(parent_type)
                parent_type.child_entities.append(entity_type)

            for parent_type in entity_type.get_all_parent_entities():
                for attr_name, other_attr in parent_type.attributes.items():
                    if attr_name not in add_attributes:
                        add_attributes[attr_name] = other_attr
                    else:
                        # allow compatible attributes
                        my_attr = add_attributes[attr_name]

                        if my_attr.type == other_attr.type:
                            add_attributes[attr_name] = other_attr
                        else:
                            raise DuplicateException(
                                my_attr, other_attr, "Incompatible attributes")
            # verify all attribute compatibility
        except TypeNotFoundException as e:
            e.set_statement(self)
            raise e


class DefineImplementation(TypeDefinitionStatement):
    """
        Define a new implementation that has a name and contains statements

        @param name: The name of the implementation
    """

    def __init__(self,
                 namespace: Namespace,
                 name: LocatableString,
                 target_type: LocatableString,
                 statements: BasicBlock,
                 comment: LocatableString):
        TypeDefinitionStatement.__init__(self, namespace, str(name))
        self.name = str(name)
        self.block = statements
        self.entity = str(target_type)

        if comment is not None:
            comment = str(comment)

        self.comment = comment

        self.type = Implementation(str(self.name), self.block, self.namespace, str(target_type), comment)
        self.type.location = name.get_location()
        self.anchors = [TypeReferenceAnchor(target_type.get_location(), namespace, str(target_type))]
        self.anchors.extend(statements.get_anchors())

    def __repr__(self) -> str:
        """
            The representation of this implementation
        """
        return "Implementation(%s)" % self.name

    def evaluate(self) -> str:
        """
            Evaluate this statement in the given scope
        """
        try:
            cls = self.namespace.get_type(self.entity)
            self.type.set_type(cls)
            self.copy_location(self.type)
        except TypeNotFoundException as e:
            e.set_statement(self)
            raise e


class DefineImplementInherits(DefinitionStatement):

    def __init__(self, entity_name: LocatableString, comment: LocatableString=None):
        DefinitionStatement.__init__(self)
        self.entity = str(entity_name)
        if comment is not None:
            self.comment = str(comment)
        else:
            self.comment = None
        self.location = entity_name.get_location()
        self.anchors.append(TypeReferenceAnchor(entity_name.get_location(), entity_name.namespace, str(entity_name)))

    def __repr__(self):
        """
            Returns a representation of this class
        """
        return "ImplementParent(%s)" % (self.entity)

    def evaluate(self):
        """
            Evaluate this statement.
        """
        try:
            entity_type = self.namespace.get_type(self.entity)

            entity_type = entity_type.get_entity()

            entity_type.implements_inherits = True
        except TypeNotFoundException as e:
            e.set_statement(self)
            raise e


class DefineImplement(DefinitionStatement):
    """
        Define a new implementation for a given entity

        @param entity: The name of the entity that is implemented
        @param implementations: A list of implementations
        @param whem: A clause that determines when this implementation is "active"
    """

    def __init__(self,
                 entity_name: LocatableString,
                 implementations: List[LocatableString],
                 select: ExpressionStatement=None,
                 comment: LocatableString=None) -> None:
        DefinitionStatement.__init__(self)
        self.entity = str(entity_name)
        self.entity_location = entity_name.get_location()
        self.implementations = [str(x) for x in implementations]
        self.anchors = [TypeReferenceAnchor(x.get_location(), x.namespace, str(x)) for x in implementations]
        self.anchors.append(TypeReferenceAnchor(entity_name.get_location(), entity_name.namespace, str(entity_name)))
        self.anchors.extend(select.get_anchors())
        self.select = select
        if comment is not None:
            self.comment = str(comment)
        else:
            self.comment = None

    def __repr__(self):
        """
            Returns a representation of this class
        """
        return "Implement(%s)" % (self.entity)

    def evaluate(self):
        """
            Evaluate this statement.
        """
        try:
            entity_type = self.namespace.get_type(str(self.entity))

            entity_type = entity_type.get_entity()

            implement = Implement()
            implement.comment = self.comment
            implement.constraint = self.select
            implement.location = self.entity_location

            i = 0
            for _impl in self.implementations:
                i += 1

                # check if the implementation has the correct type
                impl_obj = self.namespace.get_type(_impl)

                if (impl_obj.entity is not None and not
                        (entity_type is impl_obj.entity or entity_type.is_parent(impl_obj.entity))):
                    raise Exception("Type mismatch: cannot use %s as implementation for "
                                    " %s because its implementing type is %s" %
                                    (impl_obj.name, entity_type, impl_obj.entity))

                # add it
                implement.implementations.append(impl_obj)

            entity_type.add_implement(implement)
        except TypeNotFoundException as e:
            e.set_statement(self)
            raise e


class DefineTypeConstraint(TypeDefinitionStatement):
    """
        Define a new data type in the configuration. This type is a constrained
        version of a the built-in datatypes

        @param name: The name of the new  type
        @param basetype: The name of the type that is "refined"
    """

    def __init__(self, namespace, name: LocatableString, basetype: LocatableString, expression):
        TypeDefinitionStatement.__init__(self, namespace, str(name))
        self.basetype = str(basetype)
        self.anchors.append(TypeReferenceAnchor(basetype.get_location(), namespace, str(basetype)))
        self.anchors.extend(expression.get_anchors())
        self.__expression = None
        self.set_expression(expression)
        self.type = ConstraintType(self.namespace, str(name))
        self.type.location = name.get_location()
        self.comment = None

    def get_expression(self):
        """
            Get the expression that constrains the basetype
        """
        return self.__expression

    def set_expression(self, expression):
        """
            Set the expression that constrains the basetype. This expression
            should reference the value that will be assign to a variable of this
            type. This variable has the same name as the type.
        """
        contains_var = False

        if hasattr(expression, "arguments"):
            # some sort of function call
            expression = Equals(expression, Literal(True))

        for var in expression.requires():
            if var == self.name or var == "self":
                contains_var = True

        if not contains_var:
            raise TypingException(self, "typedef expressions should reference the self variable")

        self.__expression = expression

    expression = property(get_expression, set_expression)

    def __repr__(self):
        """
            A representation of this definition
        """
        return "Type(%s)" % self.name

    def evaluate(self):
        """
            Evaluate this statement.
        """
        basetype = self.namespace.get_type(self.basetype)

        constraint_type = self.type

        constraint_type.comment = self.comment
        constraint_type.basetype = basetype
        constraint_type.constraint = self.expression
        self.expression.normalize()


class DefineTypeDefault(TypeDefinitionStatement):
    """
        Define a new entity that is based on an existing entity and default values for attributes.

        @param name: The name of the new type
        @param class_ctor: A constructor statement
    """

    def __init__(self, namespace, name: LocatableString, class_ctor):
        TypeDefinitionStatement.__init__(self, namespace, str(name))
        self.type = Default(self.name)
        self.ctor = class_ctor
        self.comment = None
        self.type.location = name.get_location()
        self.anchors.extend(class_ctor.get_anchors())

    def __repr__(self):
        """
            Get a representation of this default
        """
        return "Constructor(%s, %s)" % (self.name, self.ctor)

    def evaluate(self):
        """
            Evaluate this statement.
        """
        # the base class
        type_class = self.namespace.get_type(self.ctor.class_type)

        self.type.comment = self.comment

        default = self.type
        default.set_entity(type_class)

        for name, value in self.ctor.get_attributes().items():
            default.add_default(name, value)


class DefineRelation(BiStatement):
    """
        Define a relation
    """

    def __init__(self, left, right, annotations=[]):
        DefinitionStatement.__init__(self)
        # for later evaluation
        self.annotation_expression = [(ResultVariable(), exp) for exp in annotations]
        # for access to results
        self.annotations = [exp[0] for exp in self.annotation_expression]

        self.anchors.extend((y for x in annotations for y in x.get_anchors()))
        self.anchors.append(TypeReferenceAnchor(left[0].get_location(), left[0].namespace, str(left[0])))
        self.anchors.append(TypeReferenceAnchor(right[0].get_location(), right[0].namespace, str(right[0])))

        self.left = left
        self.right = right

        self.requires = None
        self.comment = None

    def __repr__(self):
        """
            The represenation of this relation
        """
        return "Relation(%s, %s)" % (self.left[0], self.right[0])

    def evaluate(self):
        """
            Add this relation to the participating ends
        """
        try:
            left = self.namespace.get_type(str(self.left[0]))
        except TypeNotFoundException as e:
            e.set_location(self.location)
            raise e

        if isinstance(left, Default):
            raise TypingException(
                self,
                "Can not define relation on a default constructor %s, use base type instead: %s " % (
                    left.name, left.get_entity().get_full_name())
            )

        if left.get_attribute_from_related(str(self.right[1])) is not None:
            raise DuplicateException(self, left.get_attribute_from_related(str(self.right[1])),
                                     ("Attribute name %s is already defined in %s, unable to define relationship")
                                     % (str(self.right[1]), left.name))

        try:
            right = self.namespace.get_type(str(self.right[0]))
        except TypeNotFoundException as e:
            e.set_location(self.location)
            raise e

        if isinstance(right, Default):
            raise TypingException(
                self,
                "Can not define relation on a default constructor %s, use base type instead: %s " % (
                    right.name, right.get_entity().get_full_name())
            )

        if right.get_attribute_from_related(str(self.left[1])) is not None:
            raise DuplicateException(self, right.get_attribute_from_related(str(self.left[1])),
                                     ("Attribute name %s is already defined in %s, unable to define relationship")
                                     % (str(self.left[1]), right.name))

        if self.left[1] is not None:
            left_end = RelationAttribute(right, left, str(self.left[1]))
            left_end.target_annotations = self.annotations
            left_end.set_multiplicity(self.left[2])
            left_end.comment = self.comment
            left_end.location = self.left[1].get_location()
        else:
            left_end = None

        if self.right[1] is not None:
            right_end = RelationAttribute(left, right, str(self.right[1]))
            right_end.source_annotations = self.annotations
            right_end.set_multiplicity(self.right[2])
            right_end.comment = self.comment
            right_end.location = self.right[1].get_location()
        else:
            right_end = None

        if left_end is not None and right_end is not None:
            left_end.end = right_end
            right_end.end = left_end

    def emit(self, resolver, queue) -> None:
        for rv, exp in self.annotation_expression:
            reqs = exp.requires_emit(resolver, queue)
            ExecutionUnit(queue, resolver, rv, reqs, exp)

    def normalize(self) -> None:
        for _, exp in self.annotation_expression:
            exp.normalize()


class DefineIndex(DefinitionStatement):
    """
        This defines an index over attributes in an entity
    """

    def __init__(self, entity_type: LocatableString, attributes):
        DefinitionStatement.__init__(self)
        self.type = str(entity_type)
        self.attributes = [str(a) for a in attributes]
        self.anchors.append(TypeReferenceAnchor(entity_type.get_location(), entity_type.namespace, str(entity_type)))
        self.anchors.extend([AttributeReferenceAnchor(x.get_location(), entity_type.namespace,
                                                      str(entity_type), str(x)) for x in attributes])

    def types(self, recursive=False):
        """
            @see Statement#types
        """
        return [("type", self.type)]

    def __repr__(self):
        return "index %s(%s)" % (self.type, ", ".join(self.attributes))

    def evaluate(self):
        """
            Add the index to the entity
        """
        entity_type = self.namespace.get_type(self.type).get_entity()

        allattributes = entity_type.get_all_attribute_names()
        for attribute in self.attributes:
            if attribute not in allattributes:
                raise NotFoundException(self, attribute, "Attribute '%s' referenced in index is not defined in entity %s" %
                                        (attribute, entity_type))
            else:
                rattribute = entity_type.get_attribute(attribute)
                if rattribute.is_optional():
                    raise IndexException(
                        self, "Index can not contain optional attributes, Attribute ' %s.%s' is optional" %
                        (attribute, entity_type))
                if rattribute.is_multi():
                    raise IndexException(
                        self, "Index can not contain list attributes, Attribute ' %s.%s' is a list" % (attribute, entity_type))

        entity_type.add_index(self.attributes)


class PluginStatement(TypeDefinitionStatement):
    """
        This statement defines a plugin function
    """

    def __init__(self, namespace, name, function_class):
        TypeDefinitionStatement.__init__(self, namespace, name)
        self._name = name
        self._function_class = function_class
        self.type = self._function_class(namespace)

    def __repr__(self):
        """
            The representation of this function
        """
        return "Function(%s)" % self._name

    def evaluate(self):
        """
            Evaluate this plugin
        """


class DefineImport(TypeDefinitionStatement):

    def __init__(self, name: LocatableString, toname: LocatableString):
        DefinitionStatement.__init__(self)
        self.name = str(name)
        self.toname = str(toname)

    def register_types(self):
        self.target = self.namespace.get_ns_from_string(self.name)
        if self.target is None:
            raise TypeNotFoundException(self.name, self.namespace)
        self.namespace.import_ns(self.toname, self)

    def evaluate(self):
        """
            Evaluate this plugin
        """



"""
    Copyright 2018 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import random
import base64
import threading
import time

import pytest
from tornado.httpclient import HTTPRequest, AsyncHTTPClient
from inmanta import config, protocol
from inmanta.util import hash_file
from inmanta.server import config as opt
import os


def make_random_file(size=0):
    """
        Generate a random file.

        :param size: If size is > 0 content is generated that is equal or more than size.
    """
    randomvalue = str(random.randint(0, 10000))
    if size > 0:
        while len(randomvalue) < size:
            randomvalue += randomvalue

    content = ("Hello world %s\n" % (randomvalue)).encode()
    hash = hash_file(content)

    body = base64.b64encode(content).decode("ascii")

    return(hash, content, body)


@pytest.mark.gen_test
def test_client_files(client):
    (hash, content, body) = make_random_file()

    # Check if the file exists
    result = yield client.stat_file(id=hash)
    assert result.code == 404

    # Create the file
    result = yield client.upload_file(id=hash, content=body)
    assert result.code == 200

    # Get the file
    result = yield client.get_file(id=hash)
    assert result.code == 200
    assert "content" in result.result
    assert result.result["content"] == body


@pytest.mark.gen_test
def test_client_files_lost(client):
    (hash, content, body) = make_random_file()

    # Get the file
    result = yield client.get_file(id=hash)
    assert result.code == 404


@pytest.mark.gen_test
def test_sync_client_files(client):
    done = []

    def do_test():
        sync_client = protocol.SyncClient("client")

        (hash, content, body) = make_random_file()

        # Check if the file exists
        result = sync_client.stat_file(id=hash)
        assert result.code == 404

        # Create the file
        result = sync_client.upload_file(id=hash, content=body)
        assert result.code == 200

        # Get the file
        result = sync_client.get_file(id=hash)
        assert result.code == 200
        assert "content" in result.result
        assert result.result["content"] == body

        done.append(True)

    thread = threading.Thread(target=do_test)
    thread.start()

    while len(done) == 0:
        time.sleep(0.01)

    thread.join()
    assert len(done) > 0


@pytest.mark.gen_test
def test_client_files_stat(client):

    file_names = []
    i = 0
    while i < 10:
        (hash, content, body) = make_random_file()
        if hash not in file_names:
            file_names.append(hash)
            result = yield client.upload_file(id=hash, content=body)
            assert result.code == 200
            i += 1

    result = yield client.stat_files(files=file_names)
    assert len(result.result["files"]) == 0

    other_files = ["testtest"]
    result = yield client.stat_files(files=file_names + other_files)
    assert len(result.result["files"]) == len(other_files)


@pytest.mark.gen_test
def test_diff(client):
    ca = "Hello world\n".encode()
    ha = hash_file(ca)
    result = yield client.upload_file(id=ha, content=base64.b64encode(ca).decode("ascii"))
    assert(result.code == 200)

    cb = "Bye bye world\n".encode()
    hb = hash_file(cb)
    result = yield client.upload_file(id=hb, content=base64.b64encode(cb).decode("ascii"))
    assert(result.code == 200)

    diff = yield client.diff(ha, hb)
    assert(diff.code == 200)
    assert(len(diff.result["diff"]) == 5)

    diff = yield client.diff(0, hb)
    assert(diff.code == 200)
    assert(len(diff.result["diff"]) == 4)

    diff = yield client.diff(ha, 0)
    assert(diff.code == 200)
    assert(len(diff.result["diff"]) == 4)


@pytest.mark.gen_test
def test_client_files_bad(client):
    (hash, content, body) = make_random_file()
    # Create the file
    result = yield client.upload_file(id=hash + "a", content=body)
    assert result.code == 400


@pytest.mark.gen_test
def test_client_files_corrupt(client):
    (hash, content, body) = make_random_file()
    # Create the file
    result = yield client.upload_file(id=hash, content=body)
    assert result.code == 200

    state_dir = opt.state_dir.get()

    file_dir = os.path.join(state_dir, "server", "files")

    file_name = os.path.join(file_dir, hash)

    with open(file_name, "wb+") as fd:
        fd.write("Haha!".encode())

    opt.server_delete_currupt_files.set("false")
    result = yield client.get_file(id=hash)
    assert result.code == 500

    result = yield client.upload_file(id=hash, content=body)
    assert result.code == 500

    opt.server_delete_currupt_files.set("true")
    result = yield client.get_file(id=hash)
    assert result.code == 500

    result = yield client.upload_file(id=hash, content=body)
    assert result.code == 200


@pytest.mark.gen_test
def test_gzip_encoding(server):
    """
        Test if the server accepts gzipped encoding and returns gzipped encoding.
    """
    (hash, content, body) = make_random_file(size=1024)

    port = config.Config.get("server_rest_transport", "port")
    url = "http://localhost:%s/api/v1/file/%s" % (port, hash)

    zipped, body = protocol.gzipped_json({"content": body})
    assert zipped

    request = HTTPRequest(url=url, method="PUT", headers={"Accept-Encoding": "gzip", "Content-Encoding": "gzip"},
                          body=body, decompress_response=True)
    client = AsyncHTTPClient()
    response = yield client.fetch(request)
    assert response.code == 200

    request = HTTPRequest(url=url, method="GET", headers={"Accept-Encoding": "gzip"}, decompress_response=True)
    client = AsyncHTTPClient()
    response = yield client.fetch(request)
    assert response.code == 200
    assert response.headers["X-Consumed-Content-Encoding"] == "gzip"
